// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "AudioSource.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

AudioSource::AudioSource(SceneItem *i) : SceneItemComponent(i) {

}

void AudioSource::updateAudioSource() {
  if(!this->ready || data == nullptr || !data->ready) return;
  
  switch(this->requestedAudioState) {
    // When the engine wants the sound to stop
    case AUDIO_SOURCE_STATE_STOPPED:
      if(this->currentAudioState == AUDIO_SOURCE_STATE_STOPPED) return;
      this->currentAudioState = AUDIO_SOURCE_STATE_STOPPED;
      alSourceStop(this->source);
      this->eventStopped.invoke();
      break;

    // When the engine wants the sound to pause
    case AUDIO_SOURCE_STATE_PAUSED:
      if(this->currentAudioState != AUDIO_SOURCE_STATE_PLAYING) return;
      this->currentAudioState = AUDIO_SOURCE_STATE_PAUSED;
      alSourcePause(this->source);
      this->eventPaused.invoke();
      break;

    // When the engine wants the sound to play...
    case AUDIO_SOURCE_STATE_PLAYING:
      switch(this->playMode) {
        // This should be playing, right now.
        case AUDIO_PLAY_MODE_ALWAYS:
          if(this->currentAudioState == AUDIO_SOURCE_STATE_PLAYING) return;
          this->currentAudioState = AUDIO_SOURCE_STATE_PLAYING;
          alSourcePlay(this->source);
          this->eventPlaying.invoke();
          break;

        case AUDIO_PLAY_MODE_UNPAUSED:
          if(this->getGame()->timeManager.isPaused) {
            if(this->currentAudioState != AUDIO_SOURCE_STATE_PLAYING) return;
            this->currentAudioState = AUDIO_SOURCE_STATE_PAUSED;
            alSourcePause(this->source);
            this->eventPaused.invoke();
          } else {
            if(this->currentAudioState == AUDIO_SOURCE_STATE_PLAYING) return;
            this->currentAudioState = AUDIO_SOURCE_STATE_PLAYING;
            alSourcePlay(this->source);
            this->eventPlaying.invoke();
          }
          break;

        default:
          assertUnreachable();
      }
      break;

    default:
      assertUnreachable();
  }
}

void AudioSource::onStart() {
  alGenSources((ALuint)1, &this->source);

  // In future these will probably be tied to the layer
  alSourcef(this->source, AL_PITCH, 1);
  alSourcef(this->source, AL_GAIN, 1);

  glm::vec3 position = this->transform->getLocalPosition();
  alSource3f(this->source, AL_POSITION, position.x, position.y, position.z);

  // Velocity is always zero for now
  alSource3f(this->source, AL_VELOCITY, 0, 0, 0);

  // Looping
  alSourcei(this->source, AL_LOOPING, this->loop);

  // Source
  if(this->data != nullptr && this->data->ready) {
    alSourcei(source, AL_BUFFER, this->data->buffer);
  }

  // Ready.
  this->ready = true;

  // Update the source now we are ready.
  this->updateAudioSource();

  // Listen for events
  this->transform->eventTransformUpdated.addListener(this, &AudioSource::onTransformUpdate);
  this->getGame()->timeManager.eventTimePaused.addListener(this, &AudioSource::onGamePause);
  this->getGame()->timeManager.eventTimeResumed.addListener(this, &AudioSource::onGameUnpause);

}

bool_t AudioSource::isLooping() {
  return this->loop;
}

void AudioSource::setLoop(bool_t loop) {
  this->loop = loop;
  if(this->ready) alSourcei(this->source, AL_LOOPING, loop);
}

AudioSourceState AudioSource::getPlayingState() {
  return this->requestedAudioState;
}

void AudioSource::setPlayingState(enum AudioSourceState state) {
  this->requestedAudioState = state;
  this->updateAudioSource();
}

AudioSourceState AudioSource::getRealState() {
  return this->currentAudioState;
}

void AudioSource::play() {
  this->setPlayingState(AUDIO_SOURCE_STATE_PLAYING);
}

void AudioSource::pause() {
  this->setPlayingState(AUDIO_SOURCE_STATE_PAUSED);
}

void AudioSource::stop() {
  this->setPlayingState(AUDIO_SOURCE_STATE_STOPPED);
}

// void AudioSource::rewind() {
//   if(this->ready && data != nullptr && data->ready) {
//     alSourceRewind(this->source);
//   }
// }

AudioData * AudioSource::getAudioData() {
  return this->data;
}

void AudioSource::setAudioData(AudioData *data) {
  // Set up source.
  if(this->ready && data != nullptr && data->ready) {
    alSourcei(source, AL_BUFFER, data->buffer);
  }
  this->data = data;
  this->updateAudioSource();
}

enum AudioSourcePlayMode AudioSource::getPlayMode() {
  return this->playMode;
}

void AudioSource::setPlayMode(enum AudioSourcePlayMode playMode) {
  this->playMode = playMode;
  this->updateAudioSource();
}

void AudioSource::onDispose() {
  this->transform->eventTransformUpdated.removeListener(this, &AudioSource::onTransformUpdate);
  this->getGame()->timeManager.eventTimePaused.removeListener(this, &AudioSource::onGamePause);
  this->getGame()->timeManager.eventTimeResumed.removeListener(this, &AudioSource::onGameUnpause);

  this->ready = false;
  alDeleteSources((ALuint)1, &this->source);
}

void AudioSource::onTransformUpdate() {
  glm::vec3 position = this->transform->getWorldPosition();
  alSource3f(this->source, AL_POSITION, position.x, position.y, position.z);
}

void AudioSource::onGamePause() {
  this->updateAudioSource();
}

void AudioSource::onGameUnpause() {
  this->updateAudioSource();
}